#!/usr/bin/env python
# -*- coding: utf-8 -*-

# UMLS SOFTWARE: https://github.com/HHS/uts-rest-api/blob/master/samples/python/Authentication.py


# import needed libraries
import json
import requests

from lxml.html import fromstring  # type: ignore  # pylint: disable=import-error
from typing import Dict


class Authentication(object):

    def __init__(self) -> None:
        self.service: str = "http://umlsks.nlm.nih.gov"
        self.uri: str = "https://utslogin.nlm.nih.gov"
        self.auth_endpoint: str = "/cas/v1/api-key"
        self.api_key: str = open("resources/programming/umls_api.txt").read()

    def gettgt(self):
        params = {'apikey': self.api_key}
        h = {"Content-type": "application/x-www-form-urlencoded",
             "Accept": "text/plain",
             "User-Agent": "python"}

        r = requests.post(self.uri + self.auth_endpoint, data=params, headers=h)
        response = fromstring(r.text)
        tgt = response.xpath('//form/@action')[0]

        return tgt

    def getst(self, tgt):
        params = {'service': self.service}
        h = {"Content-type": "application/x-www-form-urlencoded",
             "Accept": "text/plain",
             "User-Agent": "python"}

        r = requests.post(tgt, data=params, headers=h)
        st = r.text

        return st


def cui_search(cui: str) -> Dict:
    """This function takes a string containing the UMLS REST API URI, a string containing a UMLS CUI, and a string
    specifying UMLS version to query (default is set to query the most up to date version). Using this information the
    REST API is queried. Function returns a dictionary containing the results for a single query.

    Additional information can be found: https://documentation.uts.nlm.nih.gov/rest/concept/index.html.

    Args:
        cui: A string containing a UMLS CUI.

    Returns:
        A dictionary containing the results for a single query.
    """

    auth_client = Authentication()
    tgt = auth_client.gettgt()
    content_endpoint = '/rest/content/' + 'current' + '/CUI/' + str(cui)
    query = {'ticket': auth_client.getst(tgt)}

    # ping API
    r = requests.get('https://uts-ws.nlm.nih.gov' + content_endpoint, params=query)  # type: ignore
    r.encoding = 'utf-8'

    #
    # tries, r = 0, ''
    # while r == '' and tries != 20:
    #     try:
    #         r = requests.get('https://uts-ws.nlm.nih.gov' + content_endpoint, params=query)  # type: ignore
    #         r.encoding = 'utf-8'  # type: ignore
    #     except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
    #         sleep(10)
    #         tries += 1
    #     except requests.exceptions.RequestException as e:
    #         print(e)
    #         sys.exit(1)

    return json.loads(r.text)['result']  # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from .data_utils import *
from .ontology_utils import *
from .umls_api import cui_search


__all__ = ['gets_ontology_statistics', 'gets_ontology_classes', 'gets_ontology_class_labels',
           'gets_ontology_class_labels', 'gets_ontology_class_definitions', 'gets_ontology_class_synonyms',
           'gets_ontology_class_dbxrefs', 'gets_deprecated_ontology_classes', 'cui_search', 'data_frame_subsetter',
           'data_frame_supersetter', 'column_splitter', 'aggregates_column_values', 'data_frame_grouper',
           'normalizes_source_codes', 'merge_dictionaries', 'ohdsi_ananke', 'normalizes_clinical_source_codes',
           'filters_mapping_content', 'compiles_mapping_content', 'formats_mapping_evidence',
           'assigns_mapping_category', 'aggregates_mapping_results']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Ontology Utility Functions.

Interacts with OWL Tools API
* gets_ontology_classes
* gets_deprecated_ontology_classes

"""

# import needed libraries
import os
import os.path
from rdflib import Graph, Literal, Namespace, URIRef  # type: ignore
from rdflib.namespace import RDF, OWL  # type: ignore
import subprocess

from tqdm import tqdm  # type: ignore
from typing import Dict, Set, Tuple

# set up environment variables
obo = Namespace('http://purl.obolibrary.org/obo/')
oboinowl = Namespace('http://www.geneontology.org/formats/oboInOwl#')
schema = Namespace('http://www.w3.org/2001/XMLSchema#')


def gets_ontology_classes(graph: Graph, ont_id: str) -> Set:
    """Queries a knowledge graph and returns a list of all non-deprecated owl:Class objects in the graph.

    Args:
        graph: An rdflib Graph object.
        ont_id: A string containing an ontology namespace.

    Returns:
        class_list: A set of all of the classes in the graph.

    Raises:
        ValueError: If the query returns zero nodes with type owl:ObjectProperty.
    """

    print('Querying Knowledge Graph to Obtain all OWL:Class Nodes')

    # find all classes in graph
    class_list = set([x for x in graph.subjects(RDF.type, OWL.Class) if ont_id.lower() in str(x).lower()])

    if len(class_list) > 0: return class_list
    else: raise ValueError('ERROR: No classes returned from query.')


def gets_ontology_class_labels(graph: Graph, cls: Set) -> Dict:
    """Queries a knowledge graph and returns a dictionary of all owl:Class objects and their labels in the graph.

    Args:
        graph: An rdflib Graph object.
        cls: A set of current (non-deprecated) ontology class identifiers. For example:
            {URIRef('http://purl.obolibrary.org/obo/SO_0001590)}

    Returns:
        class_list: A dictionary where keys are string labels and values are ontology URIs. An example is shown below:
            {'consensus_aflp_fragment': 'http://purl.obolibrary.org/obo/SO_0001991',
             'polypeptide_magnesium_ion_contact_site': 'http://purl.obolibrary.org/obo/SO_0001098',
             '5kb_downstream_variant': 'http://purl.obolibrary.org/obo/SO_0001633',
             'enhancer_blocking_element': 'http://purl.obolibrary.org/obo/SO_0002190', ...}
    """

    print('\nQuerying Knowledge Graph to Obtain all OWL:Class Nodes and Labels')

    # find all classes in graph
    class_list = {str(x[2]).lower(): str(x[0]) for x in tqdm(graph) if x[0] in cls and 'label' in str(x[1]).lower()}

    return class_list


def gets_ontology_class_definitions(graph: Graph, cls: Set) -> Dict:
    """Queries a knowledge graph and returns a dictionary that contains all owl:Class objects and their definitions
    in the graph.

    Args:
        graph: An rdflib Graph object.
        cls: A set of current (non-deprecated) ontology class identifiers. For example:
            {URIRef('http://purl.obolibrary.org/obo/SO_0001590)}

    Returns:
        class_list: A dictionary where keys are string definitions and values are ontology URIs. An example is shown
        below:
             {'a chromosome originating in a micronucleus.': 'http://purl.obolibrary.org/obo/SO_0000825',
              'a stop codon redefined to be a new amino acid.': 'http://purl.obolibrary.org/obo/SO_0000883',
              'a gene that is silenced by rna interference.': 'http://purl.obolibrary.org/obo/SO_0001224',
              'te that exists (or existed) in nature.': 'http://purl.obolibrary.org/obo/SO_0000797', ...}
    """

    print('\nQuerying Knowledge Graph to Obtain all OWL:Class Nodes and Definitions')

    # find all classes in graph
    class_list = {str(x[2]).lower(): str(x[0]) for x in tqdm(graph) if x[0] in cls and 'IAO_0000115' in str(x[1])}

    return class_list


def gets_ontology_class_synonyms(graph: Graph, cls: Set) -> Tuple:
    """Queries a knowledge graph and returns a tuple of dictionaries. The first dictionary contains all owl:Class
    objects and their synonyms in the graph. The second dictionary contains the synonyms and their OWL types.

    Args:
        graph: An rdflib Graph object.
        cls: A set of current (non-deprecated) ontology class identifiers. For example:
            {URIRef('http://purl.obolibrary.org/obo/SO_0001590)}

    Returns:
        A tuple of dictionaries:
            synonyms: A dictionary where keys are string synonyms and values are ontology URIs. An example is shown
                below:
                    {'modified l selenocysteine': 'http://purl.obolibrary.org/obo/SO_0001402',
                    'modified l-selenocysteine': 'http://purl.obolibrary.org/obo/SO_0001402',
                    'frameshift truncation': 'http://purl.obolibrary.org/obo/SO_0001910', ...}
            synonym_type: A dictionary where keys are string synonyms and values are OWL synonym types. An example is
                shown below:
                    {'susceptibility to herpesvirus': 'hasExactSynonym', 'full upper lip': 'hasExactSynonym'}
    """

    print('\nQuerying Knowledge Graph to Obtain all OWL:Class Nodes and Synonyms')

    # find all classes in graph
    class_list = [x for x in tqdm(graph) if x[0] in cls and 'synonym' in str(x[1]).lower()]
    synonyms = {str(x[2]).lower(): str(x[0]) for x in class_list}
    synonym_type = {str(x[2]).lower(): str(x[1]).split('#')[-1] for x in class_list}

    return synonyms, synonym_type


def gets_ontology_class_dbxrefs(graph: Graph, cls: Set) -> Tuple:
    """Queries a knowledge graph and returns a dictionary that contains all owl:Class objects and their database
    cross references (dbxrefs) in the graph. This function will also include concepts that have been identified as
    exact matches. The query returns a tuple of dictionaries where the first dictionary contains the dbxrefs and
    exact matches (URIs and labels) and the second dictionary contains the dbxref/exactmatch uris and a string
    indicating the type (i.e. dbxref or exact match).

    #TODO: This is not entirely correct. Type dicts need to be updated to include ontology ID in key.

    Assumption: That none of the hasdbxref ids overlap with any of the exactmatch ids.

    Args:
        graph: An rdflib Graph object.
        cls: A set of current (non-deprecated) ontology class identifiers. For example:
            {URIRef('http://purl.obolibrary.org/obo/SO_0001590)}

    Returns:
        dbxref: A dictionary where keys are dbxref strings and values are ontology URIs. An example is shown below:
            {'loinc:LA6690-7': 'http://purl.obolibrary.org/obo/SO_1000002',
             'RNAMOD:055': 'http://purl.obolibrary.org/obo/SO_0001347',
             'RNAMOD:076': 'http://purl.obolibrary.org/obo/SO_0001368',
             'loinc:LA6700-2': 'http://purl.obolibrary.org/obo/SO_0001590', ...}
        dbxref_type: A dictionary where keys are dbxref/exact match uris and values are string indicating if the uri
            is for a dbxref or an exact match. An example is shown below:
                {
    """

    print('\nQuerying Knowledge Graph to Obtain all OWL:Class Nodes and DbXRefs')

    # dbxrefs
    dbxref_res = [x for x in tqdm(graph) if x[0] in cls and 'hasdbxref' in str(x[1]).lower()]
    dbxref_uris = {str(x[2]).lower(): str(x[0]) for x in dbxref_res}
    dbxref_type = {str(x[2]).lower(): 'DbXref' for x in dbxref_res}

    # exact match
    exact_res = [x for x in tqdm(graph) if x[0] in cls and 'exactmatch' in str(x[1]).lower()]
    exact_uris = {str(x[2]).lower(): str(x[0]) for x in exact_res}
    exact_type = {str(x[2]).lower(): 'ExactMatch' for x in exact_res}

    # combine dictionaries
    uris = {**dbxref_uris, **exact_uris}
    types = {**dbxref_type, **exact_type}

    return uris, types


def gets_deprecated_ontology_classes(graph: Graph, ont_id: str) -> Set:
    """Queries a knowledge graph and returns a list of all deprecated owl:Class objects in the graph.

    Args:
        graph: An rdflib Graph object.
        ont_id: A string containing an ontology namespace.

    Returns:
        class_list: A list of all of the deprecated OWL classes in the graph.

    Raises:
        ValueError: If the query returns zero nodes with type owl:Class.
    """

    print('Querying Knowledge Graph to Obtain all Deprecated OWL:Class Nodes')

    # find all deprecated classes in graph
    dep_classes = list(graph.subjects(OWL.deprecated, Literal('true', datatype=URIRef(schema + 'boolean'))))
    class_list = set([x for x in dep_classes if ont_id.lower() in str(x).lower()])

    return class_list


def gets_ontology_statistics(file_location: str, owltools_location: str = './omop2obo/libs/owltools') -> None:
    """Uses the OWL Tools API to generate summary statistics (i.e. counts of axioms, classes, object properties, and
    individuals).

    Args:
        file_location: A string that contains the file path and name of an ontology.
        owltools_location: A string pointing to the location of the owl tools library.

    Returns:
        A set of all the deprecated ontology classes.

    Raises:
        TypeError: If the file_location is not type str.
        OSError: If file_location points to a non-existent file.
        ValueError: If file_location points to an empty file.
    """

    if not isinstance(file_location, str):
        raise TypeError('ERROR: file_location must be a string')
    elif not os.path.exists(file_location):
        raise OSError('The {} file does not exist!'.format(file_location))
    elif os.stat(file_location).st_size == 0:
        raise ValueError('FILE ERROR: input file: {} is empty'.format(file_location))
    else:
        output = subprocess.check_output([os.path.abspath(owltools_location), file_location, '--info'])

    # print stats
    res = output.decode('utf-8').split('\n')[-5:]
    cls, axs, op, ind = res[0].split(':')[-1], res[3].split(':')[-1], res[2].split(':')[-1], res[1].split(':')[-1]
    sent = '\nThe knowledge graph contains {0} classes, {1} axioms, {2} object properties, and {3} individuals\n'

    print(sent.format(cls, axs, op, ind))

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# import needed libraries
import hashlib
import os
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
import re

from functools import reduce
# from nltk.corpus import stopwords  # type: ignore
from nltk.stem import WordNetLemmatizer  # type: ignore
from nltk.tokenize import RegexpTokenizer  # type: ignore
from pandas import errors
from scipy import sparse  # type: ignore
from sklearn.feature_extraction.text import TfidfVectorizer  # type: ignore
from sklearn.metrics.pairwise import linear_kernel  # type: ignore
from tqdm import tqdm  # type: ignore
from typing import Dict, List, Optional, Tuple

from omop2obo.utils import column_splitter, data_frame_subsetter, merge_dictionaries

# TODO: Update script so all ontologies in the ont list (i.e. ontology_dictionary keys) are processed in parallel.

# load stopwords into environment -- horrible workaround until Travis is fixed
# nltk.download('wordnet')
stopwords = ['i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves', 'you', "you're", "you've", "you'll", "you'd",
             'your', 'yours', 'yourself', 'yourselves', 'he', 'him', 'his', 'himself', 'she', "she's", 'her', 'hers',
             'herself', 'it', "it's", 'its', 'itself', 'they', 'them', 'their', 'theirs', 'themselves', 'what',
             'which', 'who', 'whom', 'this', 'that', "that'll", 'these', 'those', 'am', 'is', 'are', 'was', 'were',
             'be', 'been', 'being', 'have', 'has', 'had', 'having', 'do', 'does', 'did', 'doing', 'a', 'an', 'the',
             'and', 'but', 'if', 'or', 'because', 'as', 'until', 'while', 'of', 'at', 'by', 'for', 'with', 'about',
             'against', 'between', 'into', 'through', 'during', 'before', 'after', 'above', 'below', 'to', 'from',
             'up', 'down', 'in', 'out', 'on', 'off', 'over', 'under', 'again', 'further', 'then', 'once', 'here',
             'there', 'when', 'where', 'why', 'how', 'all', 'any', 'both', 'each', 'few', 'more', 'most', 'other',
             'some', 'such', 'no', 'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too', 'very', 's', 't', 'can',
             'will', 'just', 'don', "don't", 'should', "should've", 'now', 'd', 'll', 'm', 'o', 're', 've', 'y', 'ain',
             'aren', "aren't", 'couldn', "couldn't", 'didn', "didn't", 'doesn', "doesn't", 'hadn', "hadn't", 'hasn',
             "hasn't", 'haven', "haven't", 'isn', "isn't", 'ma', 'mightn', "mightn't", 'mustn', "mustn't", 'needn',
             "needn't", 'shan', "shan't", 'shouldn', "shouldn't", 'wasn', "wasn't", 'weren', "weren't", 'won', "won't",
             'wouldn', "wouldn't"]


class SimilarStringFinder(object):
    """This class is designed to facilitate the mapping of clinical concepts from the Observational Medical Outcomes
    Partnership to Open Biomedical Ontology concepts. To suggest labels, we leverage clinical labels and synonyms as
    well as ontological labels, synonyms, and definitions. These items are preprocessed and then converted into a
    TF-IDF matrix, which is used to help us calculate Cosine similarity between all clinical concepts and ontology
    classes. When aligning, the user is able to pass a percentile threshold to reduce the potential list of potential
    matches and thus we return the n most similar annotations for each clinical concept. To ensure that these
    annotations are useful, we also return the similarity score for each match.

    Attributes:
        clinical_data: A Pandas DataFrame containing clinical data.
        ontology_dictionary: A nested dictionary containing ontology data, where outer keys are ontology identifiers
            (e.g. "hp", "mondo"), inner keys are data types (e.g. "label", "definition", "dbxref", and "synonyms").
            For each inner key, there is a third dictionary keyed by a string of that item type and with values that
            are the ontology URI for that string type.
        primary_key: A string containing the column name of the primary key.
        concept_strings: A list of column names containing concept-level labels and synonyms (optional).
        matrix: A Scipy sparse matrix containing the TF-IDF results for all clinical data (i.e. labels and synonyms)
            and ontology data (i.e. labels, definitions, and synonyms).

    Raises:
        TypeError:
            If clinical_file is not type str or if clinical_file is empty.
            If ontology_dictionary is not type dict.
            if primary_key is not type str.
            if concept_strings and ancestor_strings (if provided) are not type list.
        OSError:
            If the clinical_file does not exist.
    """

    def __init__(self, clinical_file: str, ontology_dictionary: Dict, primary_key: str, concept_strings: Tuple) -> None:

        self.matrix: sparse.csr_matrix = sparse.csr_matrix(0, dtype=np.int8)

        # clinical_file
        if not isinstance(clinical_file, str):
            raise TypeError('clinical_file must be type str.')
        elif not os.path.exists(clinical_file):
            raise OSError('The {} file does not exist!'.format(clinical_file))
        elif os.stat(clinical_file).st_size == 0:
            raise TypeError('Input file: {} is empty'.format(clinical_file))
        else:
            print('Loading Clinical Data')
            try:
                self.clinical_data: pd.DataFrame = pd.read_csv(clinical_file, header=0, low_memory=False).astype(str)
            except pd.errors.ParserError:
                self.clinical_data = pd.read_csv(clinical_file, header=0, sep='\t', low_memory=False).astype(str)

            self.clinical_data.fillna('', inplace=True)
            self.clinical_data = self.clinical_data.replace('nan', '')

        # check primary key
        if not isinstance(primary_key, str): raise TypeError('primary_key must be type str.')
        else: self.primary_key: str = primary_key

        # check concept-level string input (optional)
        if concept_strings is None:
            self.concept_strings: Optional[List] = concept_strings
        else:
            if not isinstance(concept_strings, Tuple):  # type: ignore
                raise TypeError('concept_strings must be type tuple.')
            else:
                self.concept_strings = list(concept_strings)

        # check ontology_dictionary
        if not isinstance(ontology_dictionary, Dict):
            raise TypeError('ontology_dictionary must be type dict.')
        else:
            self.ont_dict: Dict = ontology_dictionary

    @staticmethod
    def text_preprocessor(data: pd.DataFrame, primary_key: str) -> List:
        """Takes a Pandas DataFrame as input and performs several text preprocessing steps on one of the input columns.

        Args:
            data: A Pandas DataFrame containing two columns: (1) primary_key and (2) the column containing text to
                perform preprocessing on.
            primary_key: A string containing the name of the column to use as a primary key.

        Returns:
            data_corpus: A list of tuples, where each tuple contains an identifier and a string process.
        """

        col = [x for x in data.columns if x != primary_key][0]  # get text to perform preprocessing on
        # stop_word_patterns = re.compile(r'\b(' + r'|'.join(stopwords.words('english')) + r')\b\s*')
        stop_word_patterns = re.compile(r'\b(' + r'|'.join(stopwords) + r')\b\s*')
        lemmatizer = WordNetLemmatizer()

        # process data
        data['CODE_PROC'] = data[col].apply(lambda x: re.sub(r"\s+", " ", str(x.encode('ascii', 'ignore').decode())))
        data['CODE_PROC'] = data['CODE_PROC'].apply(lambda x: stop_word_patterns.sub('', x).lower())
        data['CODE_PROC'] = data['CODE_PROC'].apply(lambda x: RegexpTokenizer(r"\w+").tokenize(x))
        data['CODE_PROC'] = data['CODE_PROC'].apply(lambda x: [str(lemmatizer.lemmatize(i)) for i in x])

        # split column and update identifier to be primary_key_hash(string)
        hashed_str = data['CODE_PROC'].apply(lambda x: hashlib.md5(bytes(' '.join(x), 'utf-8')).hexdigest())
        data[primary_key] = data[primary_key] + '_' + hashed_str

        # convert to corpus format
        data_corpus = list(map(lambda x, y: (x, y), list(data[primary_key]), list(data['CODE_PROC'])))

        return data_corpus

    @staticmethod
    def corpus_modifier(corpus: List, ont_type: List) -> Tuple:
        """Takes a corpus, represented as a list of tuples and creates two dictionaries from it, which have been
        designed to speed up matrix-based similarity look-up.

        Args:
            corpus: A list of tuples representing a corpus, where the first item in the tuple is an identifier and
                the second item is a list of tokens. For example:
            ont_type: A list of strings specifying the ontology types (e.g. "HP", "MONDO").

        Returns:
            A tuple of dictionaries:
                (1) corpus_idx: keys are identifiers and values are lists of row identifiers.
                (2) corpus_enum: keys are row identifiers and values are indices of the row identifiers in the corpus.
        """

        # convert corpus to dictionary for faster look-up of clinical keys
        corpus_idx: Dict = {}
        for x in corpus:
            key = x[0].split('_')[0] if x[0].split('_')[0] not in ont_type else '_'.join(x[0].split('_')[0:2])
            if key in corpus_idx.keys(): corpus_idx[key].append(x[0])
            else: corpus_idx[key] = [x[0]]

        # create dict with id as key and index as value
        corpus_enum: Dict = {}
        for x, y in enumerate(corpus):
            if y[0] in corpus_enum.keys(): corpus_enum[y[0]].append(x)
            else: corpus_enum[y[0]] = [x]

        return corpus_idx, corpus_enum

    @staticmethod
    def filters_matches(matches: List, threshold: int) -> List:
        """Takes a nested list of matches, which are output from the similarity search and a threshold value and
        returns a modified list, such that only those match scores greater than or equal to the input threshold
        percentile.

        Args:
            matches: A nested list of matches from similarity search (e.g. [[0.518615172821148, 'HP_0012384']...]]).
            threshold: An integer specifying a percentile for deriving a threshold cut-off.

        Returns:
            final_matches: A list of the filtered matches (e.g. [[1.0, 'HP_0012384'], [0.7654, 'HP_00123678']]).
        """

        # filter matches to reduce duplicate results
        filtered_matches: List = []
        for x, y in sorted(matches, reverse=True):
            if not any(i for i in filtered_matches if y in i):
                filtered_matches.append([x, y])

        # derive threshold cut-off to reduce match list
        score_threshold = np.percentile([x[0] for x in filtered_matches], threshold)
        final_matches = [[str(round(x[0], 3)), x[1]] for x in filtered_matches if x[0] >= score_threshold]

        return final_matches

    def similarity_search(self, ontology_matrix: sparse.csr_matrix, var_idx: int, top_n: int) -> List:
        """The function calculates the cosine similarity between the index variables and all other included variables in
        the matrix. The results are sorted and returned as a list of lists, where each list contains a variable
        identifier and the cosine similarity score for the top set of similar variables as indicated by the input
        argument are returned.

        Args:
            ontology_matrix: A sparse Scipy matrix where each row represents a variables and each column represents a
                concept and counts are weighted by TF-IDF. This matrix only contains ontology data.
            var_idx: An integer representing a variable identifier.
            top_n: An integer representing the number of similar variables to return.

        Returns:
            similar_variables: A nested list where each list contains a variable identifier and the cosine similarity
                scores for the top_n set of similar words.
        """

        var_vector = self.matrix[var_idx:var_idx + 1]
        cosine_similarities = linear_kernel(var_vector, ontology_matrix).flatten()
        rel_var_indices = cosine_similarities.argsort()[::-1]
        similar_variables = [(variable, cosine_similarities[variable]) for variable in rel_var_indices][0:top_n]

        return similar_variables

    def scores_tfidf(self, corpus: List, ontology_type: List, top_n: int, threshold: int) -> pd.DataFrame:
        """The function iterates over the corpus and returns the top_n (as specified by user) most similar variables,
        by cosine similarity score, which are filtered to only include the top x% most similar matches.

        Args:
            corpus: A list of lists, where the first item in each list is the identifier and the second item is a list
                containing the processed question definition.
            ontology_type: A list containing ontology types (e.g. ["hp", "mondo"]) to use when filtering results.
            top_n: The number of results to return for each variable.
            threshold: An integer specifying a percentile for deriving a threshold cut-off.

        Returns:
            results: A pandas DataFrame of the top_n (as specified by user) results for each variable, with two new
                columns added per ontology. See example below:

                OUTPUT
                      CONCEPT_ID     HP_SIM_ONT_URI  HP_SIM_ONT_LABEL   MONDO_SIM_ONT_URI   MONDO_SIM_ONT_LABEL
                    0    4311399         HP_0002617        dilatation       MONDO_0001273             megacolon
        """

        onts, ont_uri = ontology_type, 'http://purl.obolibrary.org/obo/'
        ont_labels = merge_dictionaries(self.ont_dict, 'label', reverse=True)
        corpus_id, corpus_idx = self.corpus_modifier(corpus, onts)  # convert corpus to dictionary for faster look-up
        results: Dict = {x: [] for x in onts}

        # create ont-only version of TF-IDF matrix and corpus for faster cosine similarity look-up
        sub_idx = [i for j in [v for k, v in corpus_idx.items() if any(k.startswith(x) for x in onts)] for i in j]
        ont_matrix, ont_corpus = self.matrix.tocsr()[np.array(sub_idx), :], [corpus[x] for x in sub_idx]

        # matching data in filtered file
        for index, row in tqdm(self.clinical_data.iterrows(), total=self.clinical_data.shape[0]):
            match_vars = corpus_id[row[self.primary_key]]
            var_id = set([i for j in [corpus_idx[x] for x in match_vars] for i in j])
            scores = [x for y in [self.similarity_search(ont_matrix, v, top_n) for v in set(var_id)] for x in y]
            match_info = [[x[1], '_'.join(ont_corpus[x[0]][0].split('_')[0:2])] for x in scores if x[1] >= 0.25]

            if len(match_info) > 0:  # extract matches by ontology type
                for ont in onts:
                    ont_matches = [x for x in match_info if ont in x[1]]
                    if len(ont_matches) > 0:
                        hits = self.filters_matches(ont_matches, threshold)
                        results[ont].append([row[self.primary_key],
                                             ' | '.join([x[1] for x in hits]),
                                             ' | '.join([ont_labels[ont_uri + x[1]] if ont_uri + x[1] in ont_labels
                                                         else x[1] for x in hits]),
                                             ' | '.join([x[1] + '_' + x[0] for x in hits])])
                    else: continue
            else:
                continue

        # convert results to Pandas DataFrame + merge ontology findings together
        ont_sim = [pd.DataFrame({self.primary_key: [x[0] for x in results[ont]],
                                 ont + '_SIM_ONT_URI': [x[1] for x in results[ont]],
                                 ont + '_SIM_ONT_LABEL': [x[2] for x in results[ont]],
                                 ont + '_SIM_ONT_EVIDENCE': [x[3] for x in results[ont]]})
                   for ont in results.keys()]

        scored = reduce(lambda x, y: pd.merge(x, y, how='outer', on=self.primary_key), ont_sim)

        return scored.drop_duplicates()

    def performs_similarity_search(self):
        """

        Returns:
            complete_mapping: A Pandas DataFrame containing the results of calculating pairwise cosine similarity
                between the input clinical data and the input ontologies and merged back with the original input
                clinical data.
        """

        print('\n#### PERFORMING STRING SIMILARITY SEARCH ####')

        # STEP 1 - Pre-process Clinical and Ontology Data
        print('\n*** Pre-Processing Input Data ...')
        # subset input clinical data set to only include string columns
        print('Clinical Data')
        subset_data = self.clinical_data.copy()[[self.primary_key] + self.concept_strings].drop_duplicates()
        split_strings = column_splitter(subset_data, self.primary_key, self.concept_strings, '|')
        str_stacked = data_frame_subsetter(split_strings[[self.primary_key] + self.concept_strings],
                                           self.primary_key, self.concept_strings)[[self.primary_key, 'CODE']]
        preprocessed_clinical_data = self.text_preprocessor(str_stacked, self.primary_key)

        # convert ont dictionary into dictionary of Pandas DataFrames keyed by ontology type
        ont_data_dict = {}
        for ont in self.ont_dict.keys():
            print('Ontology Data: {}'.format(ont.upper()))
            ont_df = pd.concat([pd.DataFrame(self.ont_dict[ont][str_col].items(), columns=['CODE', 'ONT_URI'])
                                for str_col in ['label', 'definition', 'synonym']])
            ont_df['ONT_URI'] = ont_df['ONT_URI'].apply(lambda x: x.split('/')[-1])
            ont_data_dict[ont] = self.text_preprocessor(ont_df, 'ONT_URI')

        # STEP 2 - CREATE TF-IDF MATRIX
        print('\n*** Building TF-IDF Matrix')
        corpus = preprocessed_clinical_data + [x for y in [v for k, v in ont_data_dict.items()] for x in y]
        tf = TfidfVectorizer(tokenizer=lambda x: x, preprocessor=lambda x: x, use_idf=True, norm='l2', lowercase=False,
                             ngram_range=(1, 3))
        self.matrix = tf.fit_transform([x[1] for x in corpus])

        # STEP 3 - Calculating Cosine Similarity
        print('\n*** Calculating Cosine Similarity')
        keys = self.ont_dict.keys()
        ont_lists = [list(self.ont_dict[x]['label'].values())[0].split('/')[-1].split('_')[0] for x in keys]
        cosine_sim = self.scores_tfidf(corpus, ont_lists, 10, 75)

        # merge results with clinical data
        complete_mapping = pd.merge(self.clinical_data, cosine_sim, how='left', on=self.primary_key)
        complete_mapping.columns = [x.upper() for x in complete_mapping.columns]
        complete_mapping.fillna('', inplace=True)

        return complete_mapping



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# import needed libraries
import glob
import os
import os.path
import re
import subprocess

from datetime import *
from tqdm import tqdm  # type: ignore
from typing import Dict, List

from omop2obo.utils import gets_ontology_statistics


class OntologyDownloader(object):
    """Download a list of ontologies listed in a text file.

    Attributes:
            data_path: a string containing a file path/name to a txt file storing URLs of sources to download.
            source_list: a list of URLs representing the data sources to download/process.
            data_files: the full file path and name of each downloaded data source.
            metadata: a list containing metadata information for each downloaded ontology.

    Raises:
        TypeError: If the file pointed to by data_path is not type str.
        IOError: If the file pointed to by data_path does not exist.
        TypeError: If the file pointed to by data_path is empty.
    """

    def __init__(self, data_path: str) -> None:

        # read in data source file
        if not isinstance(data_path, str):
            raise TypeError('data_path must be type str.')
        elif not os.path.exists(data_path):
            raise OSError('The {} file does not exist!'.format(data_path))
        elif os.stat(data_path).st_size == 0:
            raise TypeError('Input file: {} is empty'.format(data_path))
        else:
            self.data_path: str = data_path

        self.source_list: Dict[str, str] = {}
        self.data_files: Dict[str, str] = {}
        self.metadata: List[List[str]] = []

    def parses_resource_file(self) -> None:
        """Parses data from a file and outputs a list where each item is a line from the input text file.

        Returns:
            source_list: A dictionary, where the key is the type of data and the value is the file path or url. See
                example below: {'chemical-gomf', 'http://ctdbase.org/reports/CTD_chem_go_enriched.tsv.gz',
                                'phenotype': 'http://purl.obolibrary.org/obo/hp.owl'
                                }

        Raises:
            ValueError: If the data within the data_path file is not formatted correctly (comma-delimited) and is not of
                the right type (.owl or .obo)
        """

        with open(self.data_path, 'r') as file_name:
            for row in file_name.read().splitlines():
                if ',' in row and ('.owl' in row or '.obo' in row):
                    self.source_list[row.strip().split(',')[0]] = row.strip().split(',')[1].strip()
                else:
                    raise ValueError('ERROR: input file: {} has incorrectly formatted data ('
                                     'i.e. non-comma-delimited rows) and/or is not of the correct type (i.e. '
                                     '.owl or .obo)'.format(self.data_path))
        file_name.close()

        return None

    def downloads_data_from_url(self, owltools_location: str = './omop2obo/libs/owltools') -> None:
        """Takes a string representing a file path/name to a text file as an argument. The function assumes
        that each item in the input file list is an URL to an OWL/OBO ontology.

        For each URL, the referenced ontology is downloaded, and used as input to an OWLTools command line argument (
        https://github.com/owlcollab/owltools/wiki/Extract-Properties-Command), which facilitates the downloading of
        ontologies that are imported by the primary ontology. The function will save the downloaded ontology + imported
        ontologies.

        Args:
            owltools_location: A string pointing to the location of the owl tools library.

        Returns:
            data_files: A dictionary mapping each source identifier to the local location where it was downloaded.
                For example: {'chemical-gomf', 'resources/edge_data/chemical-gomf_CTD_chem_go_enriched.tsv',
                              'phenotype': 'resources/ontologies/hp_with_imports.owl'
                              }
        """

        # check data before download
        self.parses_resource_file()

        # set location where to write data
        file_loc = '/'.join(self.data_path.split('/')[:-1]) + '/ontologies/'
        print('\n ***Downloading Data: to "{0}" ***\n'.format(file_loc))

        # process data
        for i in tqdm(self.source_list.keys()):
            source = self.source_list[i]
            file_prefix = source.split('/')[-1].split('.')[0]
            write_loc = file_loc + file_prefix

            print('\nDownloading: {}'.format(str(file_prefix)))

            # don't re-download ontologies
            if any(x for x in os.listdir(file_loc) if re.sub('_without.*.owl', '', x) == file_prefix):
                self.data_files[i] = glob.glob(file_loc + '*' + file_prefix + '*.owl')[0]
            else:
                try:
                    subprocess.check_call([os.path.abspath(owltools_location),
                                           str(source),
                                           '-o',
                                           str(write_loc) + '_without_imports.owl'])

                    self.data_files[i] = str(write_loc) + '_without_imports.owl'
                except subprocess.CalledProcessError as error:
                    print(error.output)

            # print stats
            gets_ontology_statistics(file_loc + str(file_prefix) + '_without_imports.owl',
                                     os.path.abspath(owltools_location))

        # generate metadata
        self.generates_source_metadata()

        return None

    def generates_source_metadata(self):
        """Obtain metadata for each imported ontology."""

        print('\n*** Generating Metadata ***\n')
        self.metadata.append(['#' + str(datetime.utcnow().strftime('%a %b %d %X UTC %Y')) + ' \n'])

        for i in tqdm(self.data_files.keys()):
            source = self.data_files[i]
            source_metadata = ['DOWNLOAD_URL= %s' % str(self.source_list[i].split(', ')[-1]),
                               'DOWNLOAD_DATE= %s' % str(datetime.now().strftime('%m/%d/%Y')),
                               'FILE_SIZE_IN_BYTES= %s' % str(os.stat(source).st_size),
                               'DOWNLOADED_FILE_LOCATION= %s' % str(source)]

            self.metadata.append(source_metadata)

        # write metadata
        self._writes_source_metadata()

        return None

    def _writes_source_metadata(self):
        """Store metadata for imported ontologies."""

        print('\n*** Writing Metadata ***\n')

        # open file to write to and specify output location
        write_loc_part = str('/'.join(list(self.data_files.values())[0].split('/')[:-1]) + '/')
        outfile = open(write_loc_part + 'ontology_source_metadata.txt', 'w')
        outfile.write('=' * 35 + '\n{}'.format(self.metadata[0][0]) + '=' * 35 + '\n\n')

        for i in tqdm(range(1, len(self.data_files.keys()) + 1)):
            outfile.write(str(self.metadata[i][0]) + '\n')
            outfile.write(str(self.metadata[i][1]) + '\n')
            outfile.write(str(self.metadata[i][2]) + '\n')
            outfile.write(str(self.metadata[i][3]) + '\n')
            outfile.write('\n')

        outfile.close()

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-


""" A Library for mapping OMOP concepts to OBO

OMOP2OBO is the first health system-wide, disease-agnostic mappings between standardized clinical terminologies in
the [Observational Medical Outcomes Partnership (OMOP) common data model and several Open Biomedical Ontologies (OBO)
foundry ontologies. These mappings were validated by domain experts and their coverage was examined in several health
systems.

There is one way to run PheKnowLator:
  1. Command line via argparse (Main.py)
"""

__all__ = [
    'ConceptAnnotator',

    'OntologyDownloader',

    'OntologyInfoExtractor',

    'SimilarStringFinder'
]

from omop2obo.clinical_concept_annotator import ConceptAnnotator
from omop2obo.ontology_downloader import OntologyDownloader
from omop2obo.ontology_explorer import OntologyInfoExtractor
from omop2obo.string_similarity import SimilarStringFinder



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Data PreProcessing Utility Functions.

Pandas DataFrame manipulations
* data_frame_subsetter
* data_frame_supersetter
* column_splitter
* aggregates_column_values
* data_frame_grouper
* normalizes_source_codes

Dictionary manipulations
* merge_dictionaries

Mapping Result Aggregation
* ohdsi_ananke
* normalizes_clinical_source_codes
* filters_mapping_content
* compiles_mapping_content
* formats_mapping_evidence
* assigns_mapping_category
* aggregates_mapping_results

"""

# import needed libraries
import pandas as pd  # type: ignore
import re

from functools import reduce
from more_itertools import unique_everseen
from tqdm import tqdm  # type: ignore
from typing import Any, Callable, Dict, List, Optional, Tuple  # type: ignore

# ENVIRONMENT WARNINGS
# WARNING 1 - Pandas: disable chained assignment warning rationale:
# https://stackoverflow.com/questions/20625582/how-to-deal-with-settingwithcopywarning-in-pandas
pd.options.mode.chained_assignment = None


def data_frame_subsetter(data: pd.DataFrame, primary_key: str, subset_columns: List) -> pd.DataFrame:
    """Takes a Pandas DataFrame and subsets it such that each subset represents an original column of codes, OMOP
    concept identifiers, and a string containing the code's column name in the original DataFrame. An example of
    the input and generated output is shown below.

        INPUT:
              CONCEPT_ID CONCEPT_SOURCE_CODE  UMLS_CUI   UMLS_CODE        UMLS_SEM_TYPE
            0    4331309            22653005  C0729608    22653005  Disease or Syndrome
            1    4331310            22653011  C4075981    22653005  Disease or Syndrome

        OUTPUT:
              CONCEPT_ID                CODE          CODE_COLUMN
            0    4331309            22653005  CONCEPT_SOURCE_CODE
            1    4331309            C0729608             UMLS_CUI
            2   37018594            22653005            UMLS_CODE

    Args:
        data: A Pandas DataFrame containing several columns of clinical codes (see INPUT for an example).
        primary_key: A string containing a column to be used as a primary key.
        subset_columns: A list of columns to subset Pandas DataFrame on.

    Returns:
        subset_data_frames: A Pandas DataFrame containing stacked subsets of the original DataFrame.
    """

    # subset data
    subset_data_frames = []

    for col in subset_columns:
        subset = data[[primary_key, col]]
        subset.loc[:, 'CODE_COLUMN'] = [col] * len(subset)
        subset.columns = [primary_key, 'CODE', 'CODE_COLUMN']
        subset_data_frames.append(subset)

    # convert list to single concatenated Pandas DataFrame
    subset_data = pd.concat(subset_data_frames)

    return subset_data.drop_duplicates()


def data_frame_supersetter(data: pd.DataFrame, index: str, columns: str, values: str) -> pd.DataFrame:
    """Takes a stacked Pandas DataFrame and unstacks it according to row values specified in the index column.
    This is equivalent to converting a DataFrame in long format to wide format. An example of the input and
    generated output is shown below.

        INPUT:
              CONCEPT_ID                CODE          CODE_COLUMN
            0    4331309            22653005  CONCEPT_SOURCE_CODE
            1    4331309            C0729608             UMLS_CUI
            2   37018594            22653005            UMLS_CODE

        OUTPUT:
                 CONCEPT_ID CONCEPT_SOURCE_CODE  UMLS_CUI   UMLS_CODE
            0    4331309            22653005  C0729608    22653005
            1    4331310            22653011  C4075981    22653005

    Args:
        data: A Pandas DataFrame containing several columns of clinical codes (see INPUT for an example).
        index: A string containing a column to be used as a primary key.
        columns: A list of columns to unstack from row values into columns.
        values: A list of columns whose values will be used to populate the unstacked DataFrame.

    Returns:
        superset_data_frame: An unstacked version of the input DataFrame (see OUTPUT above for an example).
    """

    # unstack the DataFrame
    superset_data_frame = data.drop_duplicates().pivot(index=index, columns=columns, values=values)

    # reset index
    superset_data_frame.reset_index(level=0, inplace=True)
    superset_data_frame.columns.name = None

    return superset_data_frame.drop_duplicates()


def column_splitter(data: pd.DataFrame, key: str, delimited_columns: List, delimiter: str) -> pd.DataFrame:
    """Takes a Pandas DataFrame and a list of strings specifying columns in the DataFrame that may contain a delimiter
    and expands the delimited strings within each column into separate rows. The expanded data are then merged with the
    original data.

    Args:
        data: A stacked Pandas DataFrame containing output from the umls_cui_annotator method.
        key: A string containing the column of a Pandas DataFrame to use as the primary key.
        delimited_columns: A list of the column names which contain delimited data.
        delimiter: A string specifying the delimiter type.

    Returns:
        merged_split_data: A Pandas DataFrame containing the expanded data.
    """

    delimited_data = []

    for col in delimited_columns:
        subset_data = data[[key, col]]

        # expand delimited column
        split_data = subset_data[col].str.split(delimiter).apply(pd.Series, 1).stack()
        split_data.index = split_data.index.droplevel(-1)
        split_data.name = col

        # drop original delimited column and merge expanded data
        subset_data.drop(columns=[col], inplace=True)
        merged_split_data = subset_data.join(split_data)

        # clean up leading and trailing white space
        merged_split_data[col] = merged_split_data[col].apply(lambda x: x.strip())
        delimited_data.append(merged_split_data.drop_duplicates())

    # merge delimited data
    merged_delimited_data = reduce(lambda x, y: pd.merge(x, y, on=key), delimited_data)

    return merged_delimited_data.drop_duplicates()


def aggregates_column_values(data: pd.DataFrame, primary_key: str, agg_cols: List, delimiter: str) -> pd.DataFrame:
    """Takes a Pandas DataFrame, a string containing a primary key, a list of columns to aggregate, and a string
    delimiter to use when aggregating the columns. The method then loops over each column in agg_cols and performs
    the aggregation. The method joins all aggregated columns and merges it into a single Pandas DataFrame.

    Args:
        data: A Pandas DataFrame.
        primary_key: A string containing a column name to be used as a primary key.
        agg_cols: A list of columns to aggregate.
        delimiter: A string containing a delimiter to aggregate results by.

    Returns:
        merged_combo: A Pandas DataFrame that includes the primary_key column and one
            delimiter-aggregated column for each column in the agg_cols list.
    """

    # create list of aggregated GroupBy DataFrames
    combo = [data.groupby([primary_key])[col].apply(lambda x: delimiter.join(list(unique_everseen(x))))
             for col in agg_cols]

    # merge data frames by primary key and reset index
    merged_combo = reduce(lambda x, y: pd.merge(x, y, on=primary_key, how='outer'), combo)
    merged_combo.reset_index(level=0, inplace=True)

    return merged_combo


def data_frame_grouper(data: pd.DataFrame, primary_key: str, type_column: str, col_agg: Callable) -> \
        pd.DataFrame:
    """Methods takes a Pandas DataFrame as input, a primary key, and a column to group the data by and
    creates a new DataFrame that merges the individual grouped DataFrames into a single DataFrame.
    Examples of the input and output data are shown below.

    INPUT_DATA:
                   CONCEPT_ID         CONCEPT_DBXREF_ONT_URI  CONCEPT_DBXREF_ONT_TYPE         CONCEPT_DBXREF_EVIDENCE
            0         442264        http://...MONDO_0100010                     MONDO   CONCEPT_DBXREF_sctid:68172002
            2        4029098        http://...MONDO_0045014                     MONDO  CONCEPT_DBXREF_sctid:237913008
            4        4141365           http://...HP_0000964                        HP  CONCEPT_DBXREF_sctid:426768001

    OUTPUT DATA:
    Columns: ['CONCEPT_ID', 'HP_CONCEPT_DBXREF_ONT_URI', 'HP_CONCEPT_DBXREF_ONT_LABEL', 'HP_CONCEPT_DBXREF_EVIDENCE',
              'MONDO_CONCEPT_DBXREF_ONT_URI', 'MONDO_CONCEPT_DBXREF_ONT_LABEL', 'MONDO_CONCEPT_DBXREF_EVIDENCE']

    Args:
        data: A Pandas DataFrame containing data with columns that can be grouped (see INPUT above for ex).
        primary_key: A string containing the name of the column in the input DataFrame to use as a
            primary key.
        type_column: A string containing the name of the column in the input DataFrame to use for
            grouping the data (see OUTPUT above for an example).
        col_agg: A func that aggregates data within a Pandas DataFrame column.

    Returns:
        grouped_data_full: A Pandas DataFrame containing a merged version of the grouped data.
    """

    # group data by ontology type
    grouped_data = data.groupby(type_column)
    grouped_data_frames = []

    for grp in grouped_data.groups:
        temp_df = grouped_data.get_group(grp)
        temp_df.drop(type_column, axis=1, inplace=True)
        # rename columns
        updated_names = [x.replace('ONT', grp) for x in list(temp_df.columns) if x != primary_key]
        temp_df.columns = [primary_key] + updated_names
        # aggregate data
        agg_cols = [col for col in temp_df.columns if col.split('_')[-1] in ['LABEL', 'EVIDENCE', 'URI']]
        temp_df_agg = col_agg(temp_df.copy(), primary_key, agg_cols, ' | ')

        grouped_data_frames.append(temp_df_agg.drop_duplicates())

    # merge DataFrames back together
    grouped_data_full = reduce(lambda x, y: pd.merge(x, y, how='outer', on=primary_key), grouped_data_frames)

    return grouped_data_full.drop_duplicates()


def normalizes_source_codes(data: pd.DataFrame, source_code_dict: Dict) -> pd.Series:
    """Takes a Pandas DataFrame column containing source code values that need normalization and normalizes them
    using values from a pre-built dictionary (resources/mappings/source_code_vocab_map.csv). The function is designed
    to normalize identifier prefixes according to the specifications in the source_code_dict. It provides some light
    regex support for the following scenarios:
        - ICD10CM:C85.92 --> icd10:c85.92
        - http://www.snomedbrowser.com/codes/details/12132356564 --> snomed_12132356564
        - http://www.orpha.net/ordo/orphanet_1920 --> orphanet_1920

    Assumption: assumes that the column to normalize is always the 0th index.

    Args:
        data: A column from a Pandas DataFrame containing unstacked identifiers that need normalization (e.g.
            umls:c123456, http://www.snomedbrowser.com/codes/details/12132356564, rxnorm:12345).
        source_code_dict: A dictionary keyed by input prefix with values reflecting the preferred prefixes in the
            source_code_dict file. For example:
                {'snomed': 'snomed', 'snomed_ct': 'snomed', 'snomed_ct_us_2018_03_01': 'snomed'}

    Returns:
        A Pandas Series that has been normalized.
    """

    # split prefix from number in each identifier
    prefix = data[data.columns[0]].apply(
        lambda j: j.rstrip([x for x in re.split('[_:|/]', j) if x != ''][-1])[:-1] if 'http' in j and '_' in j else
        j.rstrip([x for x in re.split('[:|/]', j) if x != ''][-1])[:-1]
    )

    id_num = data[data.columns[0]].apply(
        lambda j: [x for x in re.split('[_:|/]', j) if x != ''][-1] if 'http' in j and '_' in j else
        [x for x in re.split('[:|/]', j) if x != ''][-1]
    ).str.lower()

    # normalize prefix to dictionary and clean up urls
    norm_prefix = prefix.apply(lambda j: source_code_dict[j] if j in source_code_dict.keys() else j)

    # concat normalized identifier and number back together
    updated_source_codes = norm_prefix + ':' + id_num

    return updated_source_codes


def merge_dictionaries(dictionaries: Dict, key_type: str, reverse: bool = False) -> Dict:
    """Given any number of dictionaries, shallow copy and merge into a new dict, precedence goes to key value pairs
    in latter dictionaries.

    Function from StackOverFlow Post:
        https://stackoverflow.com/questions/38987/how-do-i-merge-two-dictionaries-in-a-single-expression-in-python

    Args:
        dictionaries: A nested dictionary.
        key_type: A string containing the key of one of the inner dictionaries.
        reverse: A bool indicating whether or not the dictionaries should be reversed before merging (default=False).

    Returns:
        combined_dictionary: A dictionary object containing.
    """

    combined_dictionary: Dict = {}

    for dictionary in dictionaries.keys():
        if reverse:
            combined_dictionary.update({v: k for k, v in dictionaries[dictionary][key_type].items()})
        else:
            combined_dictionary.update(dictionaries[dictionary][key_type])

    return combined_dictionary


def ohdsi_ananke(primary_key: str, ont_keys: list, ont_data: pd.DataFrame, data1: pd.DataFrame, data2: pd.DataFrame) \
        -> pd.DataFrame:
    """Function applies logic from the OHDSIAnanake method to extend data1, which contains dbxref mappings to OMOP
    concept ids with mappings from UMLS cuis to relevant umls ontology mappings. The merged data set is returned.

    Method adapted from: https://github.com/thepanacealab/OHDSIananke

    Args:
        primary_key: A string containing the name of the primary key (i.e. CONCEPT_ID).
        ont_keys: A list of ontology type identifiers (i.e. ['hp', 'mondo']).
        ont_data: A Pandas DataFrame containing ontology dbxref information.
        data1: A stacked Pandas DataFrame containing source codes and umls cuis.
        data2: A Pandas DataFrame containing UMLS cuis and mappings to ontologies.

    Returns:
        dbxrefs: A Pandas DataFrame containing the data from data1 merged with new entries from the umls cui data (
            data2).
    """

    # convert ont_data into a format that can be merged

    col = [x for x in ont_data.columns if 'URI' in x][0]
    ont_data['CODE'] = ont_data[col].apply(lambda x: x.split('/')[-1].lower().replace('_', ':'))

    # filter umls cui data to only
    data2_filtered = data2[(data2['CODE'].apply(lambda x: x if x.split(':')[0] in ont_keys else 999) != 999)]

    # merge with filtered data with ontology data
    merged_data = data1.merge(data2_filtered, how='inner', left_on='CODE', right_on='CUI').drop_duplicates()

    # merge with ont labels
    merged_data_ont = merged_data.merge(ont_data, how='inner', left_on='CODE_y', right_on='CODE').drop_duplicates()

    # drop unneeded columns
    dbxref_col = [x for x in merged_data_ont.columns if 'DBXREF' in x][0]
    merged_data_ont = merged_data_ont[[primary_key, 'CUI', 'CODE_COLUMN', dbxref_col]]

    # update cuis column
    merged_data_ont['CUI'] = merged_data_ont['CUI'].apply(lambda x: 'umls:' + x)

    # rename columns
    merged_data_ont.columns = [primary_key, 'CODE', 'CODE_COLUMN', dbxref_col]

    return merged_data_ont


def normalizes_clinical_source_codes(dbxref_dict: Dict, source_dict: Dict):
    """Function takes two dictionaries and uses them to create a new dictionary. The first dictionary, contains ontology
    database cross references and the second contains content to normalize the identifiers contained in the first
    dictionary.

    Args:
        dbxref_dict: A dictionary of ontology identifiers and their database cross references.
        source_dict: A dictionary containing information for normalizing the prefixes of the database cross references.

        An example of each dictionary is shown below:
        - dbxref_dict = {'DbXref', 'umls:c4022862': 'DbXref', 'umls:c0008733': 'DbXref'}
        - source_dict = {'snm', 'snomed': 'snomed', 'snomed_ct': 'snomed', 'snomed_ct_us_2018_03_01': 'snomed'}

    Returns:
        normalized_prefixes: A dictionary where the keys are database cross references and the values are strings
            containing the database cross reference type and prefix from the key, separated by a star. For example:
            {'umls:c4022862': 'DbXref*umls', 'umls:c0008733': 'DbXref*umls'}
    """

    normalized_prefixes = {}

    # split prefix from number in each identifier
    for key, value in dbxref_dict.items():
        prefix, id_num = key.split(':')[0], key.split(':')[-1].lower()
        norm_prefix = source_dict[prefix] if prefix in source_dict.keys() else prefix
        normalized_prefixes[norm_prefix + ':' + id_num] = value + '*' + prefix

    return normalized_prefixes


def filters_mapping_content(exact_results: List, similarity_results: List, threshold: float) -> Tuple:
    """Parses compiled mapping results, when results exist, to determine a final aggregated mapping result.

    Args:
        exact_results: A nested list containing 3 sub-lists where sub-list[0] contains exact match uris, sub-list[1]
            contains exact match labels, and sub-list[2] contains exact match evidence.
        similarity_results: A nested list containing 3 sub-lists where sub-list[0] contains similarity uris,
            sub-list[1] contains similarity labels, and sub-list[2] contains similarity evidence.
        threshold: A float that specifies a cut-off for filtering cosine similarity results.

    Returns:
        A tuple of lists containing mapping results for a given row. The first tuple contains exact mapping results
            and the second contains similarity results. Both lists contains 3 items: uris, labels, and evidence.
    """

    exact_uri, exact_label, exact_evid = exact_results
    sim_uri, sim_label, sim_evid = similarity_results
    exact_result: Optional[List[Any]] = [None, None, None]
    sim_result: Optional[List[Any]] = [None, None, None]

    # format results
    if exact_uri:
        exact_result = [list(unique_everseen(exact_uri)),
                        list(unique_everseen(exact_label)),
                        ' | '.join(exact_evid)]
    if sim_uri:
        if any(x for x in sim_evid[0].split(' | ') if float(x.split('_')[-1]) == 1.0):
            evid_list = sim_evid[0].split(' | ')
            sim_keep = [evid_list.index(x) for x in evid_list if float(x.split('_')[-1]) == 1.0]
            uris, labels = [sim_uri[x] for x in sim_keep], [sim_label[x] for x in sim_keep]
            sim_result = [uris, labels, ' | '.join([evid_list[x] for x in sim_keep])]
        elif any(x for x in sim_evid[0].split(' | ') if float(x.split('_')[-1]) >= threshold):
            evid_list = sim_evid[0].split(' | ')
            sim_keep = [evid_list.index(x) for x in evid_list if float(x.split('_')[-1]) >= threshold]
            uris, labels = [sim_uri[x] for x in sim_keep], [sim_label[x] for x in sim_keep]
            sim_result = [uris, labels, ' | '.join([evid_list[x] for x in sim_keep])]
        else:
            sim_result = [sim_uri, sim_label, ' | '.join(sim_evid)]

    return exact_result, sim_result


def compiles_mapping_content(row: pd.Series, ont: str, threshold: float) -> Tuple:
    """Function takes a row of data from a Pandas DataFrame and processes it to return a single ontology mapping for
    the clinical concept represented by the row.

    Args:
        row: A row from a Pandas DataFrame cont
        ont: A string containing the name of an ontology (e.g. "HP", "MONDO").
        threshold: A float that specifies a cut-off for filtering cosine similarity results.

    Returns:
        A list containing mapping results for a given row. The list contains three items: uris, labels, evidence.
    """

    relevant_cols = [x for x in row.keys() if any(y for y in ['_DBXREF_' + ont, '_STR_' + ont, ont + '_SIM'] if y in x)]

    for level in ['CONCEPT', 'ANCESTOR']:
        exact_uri, exact_label, exact_evid, sim_uri, sim_label, sim_evid = ([] for _ in range(6))  # type: ignore
        for col in relevant_cols:
            if level in col and any(y for y in ['DBXREF', 'STR'] if y in col):
                if 'URI' in col and row[col] != '': exact_uri += [x.split('/')[-1] for x in row[col].split(' | ')]
                if 'LABEL' in col and row[col] != '': exact_label += [x for x in row[col].split(' | ')]
                if 'EVIDENCE' in col and row[col] != '': exact_evid += [row[col]]
            if 'SIM' in col:
                if 'URI' in col and row[col] != '': sim_uri += [x.split('/')[-1] for x in row[col].split(' | ')]
                if 'LABEL' in col and row[col] != '': sim_label += [x for x in row[col].split(' | ')]
                if 'EVIDENCE' in col and row[col] != '': sim_evid += [row[col]]
        if exact_uri: break

    # put together mapping
    if not exact_uri and not sim_uri:
        return [None] * 3, [None] * 3  # type: ignore
    else:
        return filters_mapping_content([exact_uri, exact_label, exact_evid], [sim_uri, sim_label, sim_evid], threshold)


def formats_mapping_evidence(ont_dict: dict, source_dict: Dict, result: Tuple, clin_data: Dict) -> Tuple:
    """Takes a nested dictionary of ontology attributes, a dictionary of source code prefix mapping information, a
    nested list containing aggregated mapping information, and a dictionary of clinical concept labels and synonyms.
    The function uses this information to aggregate the evidence supporting the mapping provided in the result object.

    Args:
        ont_dict: A nested dictionary containing ontology attributes.
        source_dict: A dictionary containing ontology prefixing information.
        result: A list containing mapping results for a given row. The list contains three items: uris, labels,
            evidence.
        clin_data: A dictionary keyed by column identifier with values containing data from the keyed column.

    Returns:
        A tuple. Where the first item contains evidence for exact matches and the second contains evidence for cosine
            similarity results. For example:
            ('OBO_LABEL-OMOP_CONCEPT_LABEL:abetalipoproteinemia', 'CONCEPT_SIMILARITY:HP_0008181_1.0')
    """

    dbx_evid, lab_evid, syn_evid, sim_evid = ([] for _ in range(4))  # type: ignore
    ont_label, ont_syns, ont_syntp = ont_dict['label'], ont_dict['synonym'], ont_dict['synonym_type']
    dbxref_type = normalizes_clinical_source_codes(ont_dict['dbxref_type'], source_dict)

    # sort clinical data
    if None not in result[0]:
        for x in result[0][2].split(' | '):
            lvl = x.split('_')[0]
            clin = {k: v for k, v in clin_data.items() if lvl in k}

            if 'dbxref' in x.lower():
                if x.split('_')[-1] in dbxref_type.keys():
                    prefix = dbxref_type[x.split('_')[-1]]
                else:
                    prefix = 'DbXref*' + x.split('_')[-1].split(':')[0]
                updated_prefix = 'OBO_' + prefix.split('*')[0] + '-OMOP_' + lvl + '_CODE'
                dbx_evid.append(updated_prefix + ':' + prefix.split('*')[-1] + '_' + x.split(':')[-1].replace(':', '_'))
            if 'label' in x.lower():
                lab_evid, clin_lab = [], ' | '.join([clin[x] for x in clin.keys() if 'label' in x.lower()])
                for lab in set(clin_lab.split(' | ')):
                    if lab.lower() in ont_label.keys() and ont_label[lab.lower()].split('/')[-1] in result[0][0]:
                        lab_evid.append('OBO_LABEL-OMOP_' + x.split('_')[0] + '_LABEL:' + x.split(':')[-1])
                    if lab.lower() in ont_syns.keys() and ont_syns[lab.lower()].split('/')[-1] in result[0][0]:
                        lab_evid.append('OBO_' + ont_syntp[lab.lower()] + '-OMOP_' + lvl + '_LABEL:' + x.split(':')[-1])
            if 'synonym' in x.lower():
                syn_evid, clin_syn = [], ' | '.join([clin[x] for x in clin.keys() if 'synonym' in x.lower()])
                for syn in set(clin_syn.split(' | ')):
                    if syn.lower() in ont_label.keys() and ont_label[syn.lower()].split('/')[-1] in result[0][0]:
                        syn_evid.append('OBO_LABEL-OMOP_' + x.split('_')[0] + '_SYNONYM:' + x.split(':')[-1])
                    if clin_syn.lower() in ont_syns.keys() and ont_syns[syn.lower()].split('/')[-1] in result[0][0]:
                        syn_lab = '-OMOP_' + lvl + '_SYNONYM:'
                        syn_evid.append('OBO_' + ont_syntp[syn.lower()] + syn_lab + x.split(':')[-1])
    if None not in result[1]:
        sim_evid = ['CONCEPT_SIMILARITY:' + x for x in result[1][-1].split(' | ')]

    # compile evidence
    compiled_exact = ' | '.join(list(filter(None, list(unique_everseen(dbx_evid + lab_evid + syn_evid)))))
    compiled_sim = ' | '.join(list(filter(None, list(unique_everseen(sim_evid)))))

    return compiled_exact, compiled_sim


def assigns_mapping_category(mapping_result: List, map_evidence: str) -> str:
    """Function takes a mapping result and evidence and uses it to determine the mapping category.

    Args:
        mapping_result: A list containing mapping results for a given row. The list contains three items: uris, labels,
            evidence.
        map_evidence: A string containing the compiled mapping evidence.

    Returns:
         mapping_category: A string containing the mapping category.
    """

    if 'CONCEPT_SIMILARITY:' in map_evidence:
        if len(map_evidence.split(' | ')) > 1:
            mapping_category = 'Automatic Constructor - Concept'
        else:
            mapping_category = 'Manual Exact - Concept Similarity'
    elif any(x for x in ['ANCESTOR_CODE', 'ANCESTOR_SYNONYM', 'ANCESTOR_LABEL'] if x not in map_evidence):
        if len(mapping_result[0]) > 1:
            mapping_category = 'Automatic Constructor - Concept'
        else:
            mapping_category = 'Automatic Exact - Concept'
    elif any(x for x in ['ANCESTOR_CODE', 'ANCESTOR_SYNONYM', 'ANCESTOR_LABEL'] if x in map_evidence):
        if len(mapping_result[0]) > 1:
            mapping_category = 'Automatic Constructor - Ancestor'
        else:
            mapping_category = 'Automatic Exact - Ancestor'
    else:
        mapping_category = ''

    return mapping_category


def aggregates_mapping_results(data: pd.DataFrame, onts: List, ont_data: Dict, source_codes: Dict,
                               threshold: float = 0.25) -> pd.DataFrame:
    """Function takes a Pandas Dataframe containing the results from running the OMOP2OBO exact and similarity
    mapping functions. This function takes those results and aggregates them such that a single column set of
    evidence is returned for each ontology (i.e. uris, labels, mapping category, and mapping evidence).

    #TODO: this function could (should) be parallelized

    Args:
        data: A Pandas DataFrame of mapping results from running the OMOP2OBO exact mapping and concept similarity
            pipeline.
        onts: A list of strings representing ontologies (e.g. ["hp", "mondo"]).
        ont_data: A nested dictionary of ontology data including mappings between ontology class URIs, labels,
            synonyms, definitions, and dbxrefs.
        source_codes: A dictionary containing dbxref mappings between dbxref prefixes that is designed to normalize
            prefixes to a single type.
        threshold: A float that specifies a cut-off for filtering cosine similarity results (default = 0.25).

    Return:
        A Pandas DataFrame containing the original columns with 8 additional columns per ontology, where the first
            set of 4 columns contain the exact match results and the second set of 4 columns contains the concept
            similarity results.
    """

    print('\n#### AGGREGATING AND COMPILING MAPPING RESULTS ####')
    print('Note. Until parallelized this step can up to several hours to complete for large concept sets...\n')

    # set input variables
    cols = [x.lower() for x in data.columns]
    clin_cols = [x for x in cols if (x.endswith('label') or x.endswith('nym')) and not any(y for y in onts if y in x)]

    for ont in [x.upper() for x in onts]:
        print('Processing {} Mappings'.format(ont))
        exact_mappings: List[Any] = []
        sim_mappings: List[Any] = []
        for idx, row in tqdm(data.iterrows(), total=data.shape[0]):
            ont_list = ['DBXREF_' + ont, 'STR_' + ont, ont + '_SIM']
            res = [x for x in row.keys() if row[x] != '' and any(y for y in ont_list if y in x)]
            if len(res) != 0:
                map_info = compiles_mapping_content(row, ont, threshold)
                clin_data = {x.upper(): row[x.upper()] for x in clin_cols if x.upper() in row.keys()}
                ext_evid, sim_evid = formats_mapping_evidence(ont_data[ont.lower()], source_codes, map_info, clin_data)
                # get exact mapping information
                if ext_evid != '':
                    exact_mappings.append([' | '.join(map_info[0][0]), ' | '.join(map_info[0][1]),
                                           assigns_mapping_category(map_info[0], ext_evid), ext_evid])
                else: exact_mappings.append([None] * 4)
                # get similarity information
                if sim_evid != '':
                    sim_mappings.append([' | '.join(map_info[1][0]), ' | '.join(map_info[1][1]),
                                         assigns_mapping_category(map_info[1], sim_evid), sim_evid])
                else: sim_mappings.append([None] * 4)
            else:
                exact_mappings.append([None] * 4)
                sim_mappings.append([None] * 4)

        # add aggregated mapping results back to data frame
        data['AGGREGATED_' + ont + '_URI'] = [x[0] for x in exact_mappings]
        data['AGGREGATED_' + ont + '_LABEL'] = [x[1] for x in exact_mappings]
        data['AGGREGATED_' + ont + '_MAPPING'] = [x[2] for x in exact_mappings]
        data['AGGREGATED_' + ont + '_EVIDENCE'] = [x[3] for x in exact_mappings]
        data['SIMILARITY_' + ont + '_URI'] = [x[0] for x in sim_mappings]
        data['SIMILARITY_' + ont + '_LABEL'] = [x[1] for x in sim_mappings]
        data['SIMILARITY_' + ont + '_MAPPING'] = [x[2] for x in sim_mappings]
        data['SIMILARITY_' + ont + '_EVIDENCE'] = [x[3] for x in sim_mappings]

    # shortens long text fields in original output data (otherwise Excel expands columns into additional rows)
    size_limit = 32500  # current size limit for an Excel column
    for x in data.columns:
        data[x] = data[x].apply(lambda i: i[0:size_limit] if not isinstance(i, int) and i is not None else i)

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# import needed libraries
import os
import pandas as pd  # type: ignore

from functools import reduce
from pandas import errors
from typing import Dict, List, Optional, Tuple

from omop2obo.utils import *


class ConceptAnnotator(object):
    """An annotator to map clinical codes to ontology terms. This workflow consists of four steps that are performed
    on data from a single clinical domain:
        1 - UMLS CUI and Semantic Type Annotation
        2 - Ontology DbXRef Mapping
        3 - Exact String Mapping to concept labels and/or synonyms
        4 - Similarity distance mapping

    Attributes:
        clinical_data: A Pandas DataFrame containing clinical data.
        ontology_dictionary: A nested dictionary containing ontology data, where outer keys are ontology identifiers
            (e.g. "hp", "mondo"), inner keys are data types (e.g. "label", "definition", "dbxref", and "synonyms").
            For each inner key, there is a third dictionary keyed by a string of that item type and with values that
            are the ontology URI for that string type.
        primary_key: A string containing the column name of the primary key.
        concept_codes: A list of column names containing concept-level codes (optional).
        concept_strings: A list of column names containing concept-level labels and synonyms (optional).
        ancestor_codes: A list of column names containing ancestor concept-level codes (optional).
        ancestor_strings: A list of column names containing ancestor concept-level labels and synonyms (optional).
        umls_cui_data: A Pandas DataFrame containing UMLS CUI data from MRCONSO.RRF.
        umls_tui_data: A Pandas DataFrame containing UMLS CUI data from MRSTY.RRF.
        source_code_map: A dictionary containing clinical vocabulary source code abbreviations.
        umls_double_merge: A bool specifying whether to merge UMLS SAB codes with OMOP source codes once or twice.
            Merging once will only align OMOP source codes to UMLS SAB, twice with take the CUIs from the first merge
            and merge them again with the full UMLS SAB set resulting in a larger set of matches. The default value
            is True, which means that the merge will be performed twice.

    Raises:
        TypeError:
            If clinical_file is not type str or if clinical_file is empty.
            If source_codes is not type str or if source_codes is empty.
            If ontology_dictionary is not type dict.
            If umls_mrconso_file is not type str or if umls_mrconso_file is empty.
            If umls_mrsty_file is not type str or if umls_mrsty_file is empty.
            if primary_key is not type str.
            if concept_codes, concept_strings, ancestor_codes, and ancestor_strings (if provided) are not type list.
        OSError:
            If the clinical_file does not exist.
            If the source_codes does not  exist.
            If umls_mrconso_file does not exist.
            If umls_mrsty_file does not exist.
    """

    def __init__(self, clinical_file: str, ontology_dictionary: Dict, primary_key: str, concept_codes: Tuple,
                 concept_strings: Tuple = None, ancestor_codes: Tuple = None, ancestor_strings: Tuple = None,
                 umls_mrconso_file: str = None, umls_mrsty_file: str = None, umls_expand: bool = True,
                 source_codes: str = None) -> None:

        print('#### GENERATING EXACT MATCH MAPPINGS ####')
        print('*** Setting up Environment')

        self.umls_double_merge: bool = umls_expand

        # vocabulary source code mapping -- not tested in testing file
        source_code = 'resources/mappings/source_code_vocab_map.csv' if source_codes is None else source_codes
        if not isinstance(source_code, str):
            raise TypeError('source_codes must be type str.')
        elif not os.path.exists(source_code):
            raise OSError('The {} file does not exist!'.format(source_code))
        elif os.stat(source_code).st_size == 0:
            raise TypeError('Input file: {} is empty'.format(source_code))
        else:
            print('Loading Clinical Vocabulary Abbreviations Map')
            self.source_code_map: Dict = {}
            with open(source_code, 'r') as f:
                for x in f.read().splitlines()[1:]:
                    row = x.split(',')
                    for i in row[1].split(' | '):
                        self.source_code_map[i] = row[0]
            f.close()

        # clinical_file
        if not isinstance(clinical_file, str):
            raise TypeError('clinical_file must be type str.')
        elif not os.path.exists(clinical_file):
            raise OSError('The {} file does not exist!'.format(clinical_file))
        elif os.stat(clinical_file).st_size == 0:
            raise TypeError('Input file: {} is empty'.format(clinical_file))
        else:
            print('Loading Clinical Data')
            try:
                self.clinical_data: pd.DataFrame = pd.read_csv(clinical_file, header=0, low_memory=False).astype(str)
            except pd.errors.ParserError:
                self.clinical_data = pd.read_csv(clinical_file, header=0, sep='\t', low_memory=False).astype(str)

        # check primary key
        if not isinstance(primary_key, str):
            raise TypeError('primary_key must be type str.')
        else:
            self.primary_key: str = primary_key

        # check for concept-level information
        if not isinstance(concept_codes, Tuple):  # type: ignore
            raise TypeError('concept_codes must be type tuple.')
        else:
            self.concept_codes: List = list(concept_codes)

        # check concept-level string input (optional)
        if concept_strings is None:
            self.concept_strings: Optional[List] = concept_strings
        else:
            if not isinstance(concept_strings, Tuple):  # type: ignore
                raise TypeError('concept_strings must be type tuple.')
            else:
                self.concept_strings = list(concept_strings)

        # check ancestor-level codes input (optional)
        if ancestor_codes is None:
            self.ancestor_codes: Optional[List] = ancestor_codes
        else:
            if not isinstance(ancestor_codes, Tuple):  # type: ignore
                raise TypeError('ancestor_codes must be type tuple.')
            else:
                self.ancestor_codes = list(ancestor_codes)

        # check ancestor-level strings input (optional)
        if ancestor_strings is None:
            self.ancestor_strings: Optional[List] = ancestor_strings
        else:
            if not isinstance(ancestor_strings, Tuple):  # type: ignore
                raise TypeError('ancestor_strings must be type tuple.')
            else:
                self.ancestor_strings = list(ancestor_strings)

        # check ontology_dictionary
        if not isinstance(ontology_dictionary, Dict):
            raise TypeError('ontology_dictionary must be type dict.')
        else:
            self.ont_dict: Dict = ontology_dictionary

        # check for UMLS MRCONSO file
        # assumption (line 154) currently filtering to only keep 'ENG' codes, remove this constraint if too specific
        if not umls_mrconso_file:
            self.umls_data: Optional[pd.DataFrame] = None
        else:
            if not isinstance(umls_mrconso_file, str):
                raise TypeError('umls_mrconso_file must be type str.')
            elif not os.path.exists(umls_mrconso_file):
                raise OSError('The {} file does not exist!'.format(umls_mrconso_file))
            elif os.stat(umls_mrconso_file).st_size == 0:
                raise TypeError('Input file: {} is empty'.format(umls_mrconso_file))
            else:
                print('Loading UMLS MRCONSO Data')
                headers = ['CUI', 'LANG', 'SAB', 'CODE']
                self.umls_cui_data = pd.read_csv(umls_mrconso_file, sep='|', names=headers, low_memory=False,
                                                 header=None, usecols=[0, 1, 11, 13]).drop_duplicates().astype(str)
                # light filtering and tidying
                df = self.umls_cui_data[(self.umls_cui_data.CODE != 'NOCODE') & (self.umls_cui_data.LANG == 'ENG')]
                self.umls_cui_data = df[['CUI', 'SAB', 'CODE']].drop_duplicates()
                self.umls_cui_data['CODE'] = self.umls_cui_data['SAB'] + ':' + self.umls_cui_data['CODE'].str.lower()
                self.umls_cui_data['CODE'] = self.umls_cui_data['CODE'].apply(
                    lambda j: ':'.join(j.split(':')[1:]) if len(j.split(':')) > 2 else j)
                self.umls_cui_data['CODE'] = normalizes_source_codes(self.umls_cui_data['CODE'].to_frame(),
                                                                     self.source_code_map)

        # check for UMLS MRSTY file
        if not umls_mrsty_file:
            self.umls_tui_data: Optional[pd.DataFrame] = None
        else:
            if not isinstance(umls_mrsty_file, str):
                raise TypeError('umls_mrsty_file must be type str.')
            elif not os.path.exists(umls_mrsty_file):
                raise OSError('The {} file does not exist!'.format(umls_mrsty_file))
            elif os.stat(umls_mrsty_file).st_size == 0:
                raise TypeError('Input file: {} is empty'.format(umls_mrsty_file))
            else:
                print('Loading UMLS MRSTY Data')
                headers = ['CUI', 'STY']
                self.umls_tui_data = pd.read_csv(umls_mrsty_file, header=None, sep='|', names=headers,
                                                 low_memory=False, usecols=[0, 3]).drop_duplicates().astype(str)

    def umls_cui_annotator(self, data: pd.DataFrame, key: str, code_level: str) -> pd.DataFrame:
        """Method maps concepts in a clinical data file to UMLS concepts and semantic types from the umls_cui_data
        and umls_tui_data Pandas DataFrames.

        Args:
            data: A Pandas DatFrame containing clinical data.
            key: A string containing the name of the primary key (i.e. CONCEPT_ID).
            code_level: A string containing the name of the source code column (i.e. CONCEPT_SOURCE_CODE).


        Returns:
           umls_cui_semtype: A Pandas DataFrame containing clinical concept ids and source codes as well as UMLS
            CUIs, source codes, and semantic types. An example of the output data is shown below:

                          CONCEPT_ID    CONCEPT_SOURCE_CODE     UMLS_CUI      UMLS_CODE           UMLS_SEM_TYPE
                    0        4331309               22653005     C0729608       22653005     Disease or Syndrome
                    1        4331309               22653005     C0729608       22653005     Disease or Syndrome
                    2       37018594         80251000119104     C4075981 80251000119104                 Finding

        """

        # reduce data to only those columns needed for merging
        clinical_ids = data[[key, code_level]].drop_duplicates()

        # merge reduced clinical concepts with umls concepts
        if self.umls_double_merge is True:
            # merge 1 - align omop source codes to umls sabs
            umls_cui_1 = clinical_ids.merge(self.umls_cui_data, how='inner', left_on=code_level, right_on='CODE')
            # merge 2 - align umls cuis from merge 1 to cuis in full umls (this adds additional sabs not found in omop)
            umls_cui_2 = umls_cui_1[[key, code_level, 'CUI']].merge(self.umls_cui_data, how='left', on='CUI')
            umls_cui = pd.concat([umls_cui_1, umls_cui_2])
        else:
            umls_cui = clinical_ids.merge(self.umls_cui_data, how='inner', left_on=code_level, right_on='CODE')

        umls_cui_semtype = umls_cui.merge(self.umls_tui_data, how='left', on='CUI').drop_duplicates()

        # update column names
        umls_cui_semtype.columns = [key, code_level, 'UMLS_CUI', 'UMLS_SAB', 'UMLS_CODE', 'UMLS_SEM_TYPE']

        return umls_cui_semtype

    def dbxref_mapper(self, data: pd.DataFrame, primary_key: str, code_type: str) -> pd.DataFrame:
        """Takes a stacked Pandas DataFrame and merges it with a Pandas DataFrame version of the
        ontology_dictionary_object.

            INPUT (ignore all columns starting with "UMLS" if UMLS data is not provided):
                      CONCEPT_ID            CODE          CODE_COLUMN
                0        4331309        22653005  CONCEPT_SOURCE_CODE
                1       37018594  80251000119104  CONCEPT_SOURCE_CODE
                2         442264        68172002  CONCEPT_SOURCE_CODE

            OUTPUT:
                    CONCEPT_ID      CODE          CODE_COLUMN           DBXREF    ONT   ONT_URI          EVIDENCE
                0       442264  68172002  CONCEPT_SOURCE_CODE   SCTID:68172002  MONDO       URL      DbXRef_SCTID
                1       442264  68172002            UMLS_CODE   SCTID:68172002  MONDO       URL      DbXRef_SCTID
                2      4029098 237913008  CONCEPT_SOURCE_CODE  SCTID:237913008  MONDO       URL      DbXRef_SCTID
        Args:
            data: A stacked Pandas DataFrame containing output from the umls_cui_annotator method (see INPUT above).
            primary_key: A string containing the name of the primary key (i.e. CONCEPT_ID).
            code_type: A string containing the concept_level (i.e. concept or ancestor).

        Returns:
            merged_dbxrefs: A stacked Pandas DataFrame containing ontology dbxref merging results (see OUTPUT above).
        """

        col_lab = code_type.upper() + '_DBXREF_ONT_'  # column labels
        ont_labels = merge_dictionaries(self.ont_dict, 'label', reverse=True)

        # convert ontology dictionary to Pandas DataFrame
        ont_df = pd.concat([pd.DataFrame(self.ont_dict[ont]['dbxref'].items(), columns=['CODE', col_lab + 'URI'])
                            for ont in self.ont_dict.keys() if len(self.ont_dict[ont]['dbxref']) > 0])
        # normalize source_code prefix values
        ont_df['CODE'] = normalizes_source_codes(ont_df['CODE'].to_frame(), self.source_code_map)
        # merge ontology data and clinical data and run ohdsi ananke approach to specifically pull umls ont mappings
        if self.umls_cui_data is not None:
            dbxrefs = pd.concat(
                [data.merge(ont_df, how='inner', on='CODE').drop_duplicates(),
                 ohdsi_ananke(primary_key, list(self.ont_dict.keys()), ont_df.copy(), data, self.umls_cui_data.copy())]
            )
        else:
            dbxrefs = data.merge(ont_df, how='inner', on='CODE').drop_duplicates()

        # update content and labels
        dbxrefs[col_lab + 'TYPE'] = dbxrefs[col_lab + 'URI'].apply(lambda x: x.split('/')[-1].split('_')[0])
        dbxrefs[col_lab + 'LABEL'] = dbxrefs[col_lab + 'URI'].apply(lambda x: ont_labels[x])
        # update evidence formatting --> EX: CONCEPTS_DBXREF_UMLS:C0008533
        dbxrefs[col_lab + 'EVIDENCE'] = dbxrefs['CODE'].apply(lambda x: col_lab[0:-4] + x)
        # drop unneeded columns
        dbxrefs = dbxrefs[[primary_key] + [x for x in list(dbxrefs.columns) if x.startswith(col_lab[0:-4])]]

        return dbxrefs.drop_duplicates()

    def exact_string_mapper(self, data: pd.DataFrame, primary_key: str, code_type: str) -> pd.DataFrame:
        """Takes a stacked Pandas DataFrame and merges it with a Pandas DataFrame version of the ontology-dictionary
        object 'label' and 'synonym' data.

            INPUT:
                    CONCEPT_ID                         CONCEPT_LABEL                                 CONCEPT_SYNONYM
                0      4331309   Myocarditis due to infectious agent              Myocarditis due to infectious agent
                1      4331309   Myocarditis due to infectious agent                            Infective myocarditis
                2      4331309   Myocarditis due to infectious agent   Myocarditis due to infectious agent (disorder)

            OUTPUT:
                      CONCEPT_ID                  CODE     CODE_COLUMN   ONT_URI     ONT             EVIDENCE
                0        4141365  engraftment syndrome   CONCEPT_LABEL       URL   MONDO     LABEL_CONCEPT_ID
                1        4141365  engraftment syndrome  CONCEPT_SYNONYM      URL   MONDO     LABEL_CONCEPT_ID
                2         133835                eczema    CONCEPT_LABEL      URL      HP     LABEL_CONCEPT_ID
        Args:
            data: A stacked Pandas DataFrame containing output from the umls_cui_annotator method (see INPUT above
                for an example).
            primary_key: A string containing the name of the primary key (i.e. CONCEPT_ID).
            code_type: A string containing the concept_level (i.e. concept or ancestor).

        Returns:
            merged_strings: A Pandas DataFrame containing the results from string matching the ontology strings to
                the clinical strings (see OUTPUT above for an example).
        """

        col_label = code_type.upper() + '_STR_ONT_'  # column labels
        ont_labels = merge_dictionaries(self.ont_dict, 'label', reverse=True)
        data['CODE'] = data['CODE'].apply(lambda x: x.lower())  # prepare clinical data

        # prepare ontology data
        ont_dfs = []
        for str_col in ['label', 'synonym']:
            combo_dict_df_label = pd.concat([pd.DataFrame(self.ont_dict[ont][str_col].items(),
                                                          columns=['CODE', col_label + 'URI']) for ont in
                                             self.ont_dict.keys()])

            # merge ontology data and clinical data
            str_data = data.merge(combo_dict_df_label, how='inner', on='CODE').drop_duplicates()
            # update ontology data formatting
            str_data[col_label + 'TYPE'] = str_data[col_label + 'URI'].apply(lambda x: x.split('/')[-1].split('_')[0])
            str_data[col_label + 'LABEL'] = str_data[col_label + 'URI'].apply(lambda x: ont_labels[x])
            # update evidence formatting --> EX: CONCEPT_SYNONYM:dic_in_newborn
            aggregated_evidence = str_data['CODE'].apply(lambda x: x.replace(' ', '_'))
            str_data[col_label + 'EVIDENCE'] = str_data['CODE_COLUMN'] + ':' + aggregated_evidence
            # drop unneeded columns
            str_data = str_data[[primary_key] + [x for x in list(str_data.columns) if x.startswith(col_label[0:-4])]]
            ont_dfs.append(str_data.drop_duplicates())

        return pd.concat(ont_dfs).drop_duplicates()

    def clinical_concept_mapper(self) -> pd.DataFrame:
        """This method serves as the main method for this class. it's purpose is to iterate over all relevant data in
        an input clinical data file and generate several different kinds of mappings to a ontologies provided in an
        input dictionary. The script annotates the data on two levels: concepts and concept ancestors. For both
        levels, the following steps are completed to derive concept annotations:
            1 - UMLS CUIs and semantics types to concept ids
            2 - Ontology dbXRefs between concept ids and ontology ids
            3 - Exact string matching concept labels and synonyms to ontology labels and synonyms
            4 - Aggregating the results from steps 1-3 into a single Pandas DataFrame
            5 - Combine results from each level into single Pandas DataFrame

        Returns:
            complete_map: A Pandas DataFrame containing the results of performing dbXRef and exact string mapping to
                the input ontologies and clinical data.
        """

        level_maps = []

        if self.ancestor_codes is not None:
            levels = {'concept': {'codes': self.concept_codes, 'strings': self.concept_strings},
                      'ancestor': {'codes': self.ancestor_codes,
                                   'strings': self.ancestor_strings}}
        else:
            levels = {'concept': {'codes': self.concept_codes, 'strings': self.concept_strings}}

        for level in levels.keys():
            print('\n*** Annotating Level: {}'.format(level))
            primary_key, data = self.primary_key, self.clinical_data.copy()
            code_level, code_strings = levels[level]['codes'][0], levels[level]['strings']  # type: ignore
            if level == 'ancestor' or any(x for x in data[code_level] if '|' in x):
                data = column_splitter(data, primary_key, [code_level], '|')[[primary_key] + [code_level]]
                data[code_level] = normalizes_source_codes(data[code_level].to_frame(), self.source_code_map)
            else:
                data[code_level] = normalizes_source_codes(data[code_level].to_frame(), self.source_code_map)

            # STEP 1: UMLS CUI + SEMANTIC TYPE ANNOTATION
            print('Performing UMLS CUI + Semantic Type Annotation')
            if self.umls_cui_data is not None and self.umls_tui_data is not None:
                umls_map = self.umls_cui_annotator(data.copy(), primary_key, code_level)
                sub = [code_level, 'UMLS_CODE', 'UMLS_CUI']
                data_stacked = data_frame_subsetter(umls_map[[primary_key] + sub], primary_key, sub)
            else:
                print('Did not provide MRCONSO and MRSTY Files -- Skipping UMLS Annotation Step')
                umls_map, clinical_subset = None, data[[primary_key, code_level]]
                data_stacked = data_frame_subsetter(clinical_subset, primary_key, [code_level])

            # STEP 2 - DBXREF ANNOTATION
            print('Performing DbXRef Annotation')
            stacked_dbxref = self.dbxref_mapper(data_stacked.copy(), primary_key, level)
            # files = 'resources/mappings/' + level + '_dbXRef_Mappings.csv'
            # stacked_dbxref.to_csv(files, sep=',', index=False, header=True)

            # STEP 3 - EXACT STRING MAPPING
            print('Performing Exact String Mapping')
            clinical_strings = self.clinical_data.copy()[[primary_key] + code_strings]  # type: ignore
            split_strings = column_splitter(clinical_strings, primary_key, code_strings, '|')  # type: ignore
            split_strings = split_strings[[primary_key] + code_strings]  # type: ignore
            split_strings_stacked = data_frame_subsetter(split_strings, primary_key, code_strings)  # type: ignore
            stacked_strings = self.exact_string_mapper(split_strings_stacked, primary_key, level)
            # files_str = 'resources/mappings/' + level + '_String_Mappings.csv'
            # stacked_strings.to_csv(files_str, sep=',', index=False, header=True)

            # STEP 4 - COMBINE RESULTS
            print('Aggregating Mapping Results')
            # dbXRef annotations
            if len(stacked_dbxref) != 0:
                ont_type_column = [col for col in stacked_dbxref.columns if 'TYPE' in col][0]
                dbxrefs = data_frame_grouper(stacked_dbxref.copy(), primary_key, ont_type_column,
                                             aggregates_column_values)
            else:
                dbxrefs = None

            # exact string annotations
            if len(stacked_strings) != 0:
                ont_type_column = [col for col in stacked_strings.columns if 'TYPE' in col][0]
                strings = data_frame_grouper(stacked_strings.copy(), primary_key, ont_type_column,
                                             aggregates_column_values)
            else:
                strings = None

            # umls annotations
            if umls_map is not None:
                umls, agg_cols = umls_map[[primary_key, 'UMLS_CUI', 'UMLS_SEM_TYPE']], ['UMLS_CUI', 'UMLS_SEM_TYPE']
                umls = aggregates_column_values(umls.copy(), primary_key, agg_cols, ' | ')
                umls.columns = [primary_key] + [level.upper() + '_' + x for x in umls.columns if x != primary_key]
            else:
                umls = None

            # combine annotations
            dfs = [x for x in [dbxrefs, strings, umls] if x is not None]
            if len(dfs) > 1:
                level_maps.append(reduce(lambda x, y: pd.merge(x, y, how='outer', on=primary_key), dfs))
            else:
                level_maps.append(dfs[0])

        # STEP 5 - COMBINE CONCEPT AND ANCESTOR DATA
        print('Combining Concept and Ancestor Maps')
        full_map = reduce(lambda x, y: pd.merge(x, y, how='outer', on=self.primary_key), level_maps)
        complete_map = pd.merge(self.clinical_data, full_map, how='left', on=self.primary_key)
        complete_map.columns = [x.upper() for x in complete_map.columns]
        complete_map.fillna('', inplace=True)

        return complete_map



import os
import re

# To use a consistent encoding
from codecs import open as copen
from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
with copen(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


def read(*parts):
    with copen(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


__version__ = find_version("omop2obo", "__version__.py")

test_deps = [
    'codacy-coverage',
    'coveralls',
    'mock',
    'mypy',
    'pytest',
    'pytest-cov',
    'validate_version_code'
]

extras = {
    'test': test_deps,
}

setup(
    name='omop2obo',
    version=__version__,
    description="OMOP2OBO is the first health system-wide, disease-agnostic mappings between standardized clinical "
                "terminologies in the Observational Medical Outcomes Partnership (OMOP) common data model and several "
                "Open Biomedical Ontologies (OBO).",
    long_description=long_description,
    url="https://github.com/callahantiff/OMOP2OBO",
    keywords='omop, omop-cdm, formal-logic, biomedical-applications, '
             'open-biomedical-ontologies, obofoundry, clinical-terminologies, ',
    author='Tiffany J. Callahan',
    author_email='tiffany.callahan@cuanschutz.edu',

    # Choose your license
    license='MIT',
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    tests_require=test_deps,

    # Add here the package dependencies
    install_requires=[
        'click==8.0.0a1',
        'lxml==4.6.2',
        'more-itertools==8.6.0',
        'nltk==3.6.5',
        'oauth2client==4.1.3',
        'openpyxl==3.0.5',
        'pandas==1.1.5',
        'rdflib==5.0.0',
        'regex>=2021.8.3',
        'responses==0.10.12',
        'scipy==1.5.4',
        'scikit-learn==0.23.2',
        'tqdm==4.54.1',
        'types-requests'],
    extras_require=extras,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# import needed libraries
import glob
import os
import pickle

from rdflib import Graph  # type: ignore
from tqdm import tqdm  # type: ignore
from typing import Dict

from omop2obo.utils import *


class OntologyInfoExtractor(object):
    """Class creates an RDF graph from an OWL file and then performs queries to return DbXRefs, synonyms, and labels.

    Attributes:
        graph: An rdflib graph object.
        ont_dictionary: A dictionary, where keys are a string aliasing an ontology and values are a string pointing
            the local location where the ontology was downloaded.
        ont_directory: A string containing the filepath to the ontology data directory.

    Raises:
        OSError: If the ontology_dictionary cannot be found.
        TypeError: If ontology_dictionary is empty.
    """

    def __init__(self, ontology_directory: str, ont_dictionary: Dict) -> None:
        self.graph: Graph = Graph()
        self.ont_dictionary = ont_dictionary

        # check for ontology data
        if not os.path.exists(ontology_directory):
            raise OSError("Can't find 'ontologies/' directory, this directory is a required input")
        elif len(glob.glob(ontology_directory + '/*.owl')) == 0:
            raise TypeError('The ontologies directory is empty')
        else:
            self.ont_directory = ontology_directory

    def get_ontology_information(self, ont_id: str) -> Dict:
        """Function queries an RDF graph and returns labels, definitions, dbXRefs, and synonyms for all
        non-deprecated ontology classes.

        Args:
            ont_id: A string containing an ontology namespace.

        Returns: A dict mapping each DbXRef to a list containing the corresponding class ID and label. For example:
            {'HP': {
                'label': {'narrow naris': 'http://purl.obolibrary.org/obo/HP_0009933'},
                'definition': {'agenesis of lower primary incisor.': 'http://purl.obolibrary.org/obo/HP_0011047'},
                'dbxref': {'SNOMEDCT_US:88598008': 'http://purl.obolibrary.org/obo/HP_0000735'},
                'synonyms': { 'open bite': 'http://purl.obolibrary.org/obo/HP_0010807'}}
        """

        # get class information
        deprecated_classes = gets_deprecated_ontology_classes(self.graph, ont_id)
        filter_classes = set([x for x in gets_ontology_classes(self.graph, ont_id) if x not in deprecated_classes])

        # filter results and add to dictionary
        syns = gets_ontology_class_synonyms(self.graph, filter_classes)
        dbxrefs = gets_ontology_class_dbxrefs(self.graph, filter_classes)
        res: Dict = {
            'label': gets_ontology_class_labels(self.graph, filter_classes),
            'definition': gets_ontology_class_definitions(self.graph, filter_classes),
            'dbxref': dbxrefs[0], 'dbxref_type': dbxrefs[1],
            'synonym': syns[0], 'synonym_type': syns[1],
        }

        return res

    def ontology_processor(self) -> None:
        """Using different information from the user, this function retrieves all class labels, definitions,
        synonyms, and database cross-references (dbXref). The function expects a dictionary as input where the keys are
        short nick-names or OBO abbreviations for ontologies and the values are lists, where the first item is a string
        that contains the file path information to the downloaded ontology, the second item is a list of clinical
        identifiers that can be used for filtering the dbXrefs. An example of this input is shown below.

            {'CHEBI': ['resources/ontologies/chebi_without_imports.owl', ['DrugBank', 'ChEMBL', 'UniProt']]}

        Returns:
            None.
        """

        for ont in self.ont_dictionary.items():
            print('\nPROCESSING ONTOLOGY: {0}'.format(ont[0]))

            # check and make sure that we are not re-running a file we have already processed
            if ont[1].replace('.owl', '_class_information.pickle').split('/')[-1] in os.listdir(self.ont_directory):
                pass
            else:
                print('Loading RDF Graph ... Please be patient, this step can take several minutes for large files.')
                self.graph = Graph().parse(ont[1], format='xml')

                # get ontology information
                ont_dict = self.get_ontology_information(ont[0])

                with open(str(ont[1][:-4]) + '_class_information.pickle', 'wb') as handle:
                    pickle.dump(ont_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)
                handle.close()

        return None

    def ontology_loader(self) -> None:
        """Function takes a list of file paths to pickled data, loads the data, and then saves each file as a dictionary
        entry.

        Returns:
            A dictionary where each key is a file name and each value is a dictionary.

        Raises:
            OSError: If there are no pickled data files in the directory specified by the ont_directory variable.
        """

        # find pickled data
        pickled_data = glob.glob(self.ont_directory + '/*_class_information.pickle')

        if len(pickled_data) == 0:
            raise OSError('There are no ontology pickled data files in: {}'.format(self.ont_directory))
        else:
            ontology_data = {}

            for ont_file in tqdm(pickled_data):
                if 'master' not in ont_file:  # ignore existing pickled master_ontology_dictionary files
                    with open(ont_file, 'rb') as _file:
                        ont_data = pickle.load(_file)
                        ont_id = list(ont_data['label'].values())[0].split('/')[-1].split('_')[0].lower()
                        ontology_data[ont_id] = ont_data
                    _file.close()

            # write out all ontologies into a single pickled file
            with open(self.ont_directory + '/master_ontology_dictionary.pickle', 'wb') as handle:
                pickle.dump(ontology_data, handle, protocol=pickle.HIGHEST_PROTOCOL)
            handle.close()

            return None
